#include <stdio.h>
#include <limits.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/uio.h>
#include <net/if.h>
#include <net/if_tun.h>
#include <netinet/in.h>
#include <netinet/ip6.h>
#include <netdb.h>
#include <poll.h>
#include <errno.h>

extern int h_errno;

#define UDP6TO4PORT   0x2003
#define UDP6TO4PREFIX 0x2003

struct tuninfo {
  int debug;
  int tunnumber;
  int tunfd;
  int udpfd;
  int meserver;
  struct in_addr echoserver;
  struct in_addr myaddr;
  int myaddr_set;
  struct hostent *he;
};

tuninit(struct tuninfo *ti)  
{
  char tunname[_POSIX_PATH_MAX];
  int tunctrl;
  int tunmode;
  
  snprintf(tunname, _POSIX_PATH_MAX, "/dev/tun%d", ti->tunnumber);

  if((tunctrl=open(tunname, O_RDWR))<0) {
    perror(tunname);
    exit(1);
  }

  ti->tunfd = tunctrl;

  tunmode = IFF_BROADCAST;

  if(ioctl(ti->tunfd, TUNSIFMODE, &tunmode) == -1) {
    perror("tunsifmode");
    exit(2);
  }
}

void udpinit(struct tuninfo *ti)
{
  struct sockaddr_in sudp;
  int rc;

  bzero(&sudp, sizeof(struct sockaddr_in));
  sudp.sin_family=AF_INET;
  sudp.sin_len = sizeof(struct sockaddr_in);
  sudp.sin_addr.s_addr = htonl(INADDR_ANY);
  sudp.sin_port = htons(UDP6TO4PORT);

  ti->udpfd = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
  if(ti->udpfd == -1) {
    perror("socket");
    exit(1);
  }

  if(bind(ti->udpfd, (struct sockaddr *)&sudp, sudp.sin_len) == -1) {
    perror("bind");
    exit(1);
  }

  rc = 1;

  if (setsockopt(ti->udpfd, IPPROTO_IP, IP_RECVDSTADDR, &rc, sizeof(rc))==-1)
    errx(1, "setsockopt IP_RECVDSTADDR: %s", strerror(errno));

}

void send_echorequest(struct tuninfo *ti)
{
  struct sockaddr_in dudp;
  char datagram[5];

  /* now, send a request to remote server for an echo response */
  /* the response, if any, will be processed by the main loop, and
   * will cause the tunnel device to be appropriately configured. */

  datagram[0]=0x11; /* an invalid IPv4 version, with too little data */
  
  bzero(&dudp, sizeof(struct sockaddr_in));
  dudp.sin_family=AF_INET;
  dudp.sin_len = sizeof(struct sockaddr_in);
  dudp.sin_addr = ti->echoserver;
  dudp.sin_port = htons(UDP6TO4PORT);

  if(sendto(ti->udpfd, datagram, 1, 0,
	    (struct sockaddr *)&dudp, dudp.sin_len) == -1) {
    perror("sendto");
  }
}

tunprocess(struct tuninfo *ti)
{
  unsigned char buf[2048];
  int readlen;
  struct ip6_hdr *ip6;
  char addrbuf[64];
  struct sockaddr_in dst;
  struct in6_addr *dst6;
  
  readlen = read(ti->tunfd, buf, 2048);
  if(readlen == -1) {
    perror("read from tunnel");
    exit(1);
  }

  /* now take the IPv6 address apart, and send it out accordingly */
  ip6 = (struct ip6_hdr *)buf;

  if((ip6->ip6_vfc & IPV6_VERSION_MASK) != IPV6_VERSION) {
    fprintf(stderr, "Got strange packet... version byte is %02x\n",
	    ip6->ip6_vfc);
    return;
  }

  dst6 = &ip6->ip6_dst;

  /* okay, look at the address now */
  if(dst6->s6_addr[0] != 0x20 ||
     dst6->s6_addr[1] != 0x03) {

    /* oops, this isn't a 6to4UDP address. Better grab the default route
       instead */
    /*    ip6 = ti->default_addr; */

    fprintf(stderr, "Non 0x2003 address received %s\n",
	    inet_ntop(AF_INET6, dst6, addrbuf, sizeof(addrbuf)));
    exit(5);
  }

  /* now look at rest of address to form address */
  bzero(&dst, sizeof(dst));
  dst.sin_family = AF_INET;
  dst.sin_len    = sizeof(struct sockaddr_in);
  bcopy(&dst6->s6_addr[2], &dst.sin_addr, 4);
  bcopy(&dst6->s6_addr[6], &dst.sin_port, 2);
  
  if(sendto(ti->udpfd, buf, readlen, 0,
	    (struct sockaddr *)&dst, dst.sin_len) == -1) {
    perror("UDP v6 sendto");
  }

  return;
}

udpprocess(struct tuninfo *ti)
{
  unsigned char buf[2048];
  struct msghdr mh;
  struct sockaddr_in addrs[2];
  struct sockaddr_in to, from, translatedaddr;
  unsigned short translatedport;
  int msglen;
  struct iovec iov[1];
  char   addrfrom[32], addrto[32];
  char   sysbuf[256];
  unsigned char *ipdst;
  struct cmsghdr *cmsg;
  unsigned char  address[16384];
  int tovalid;

  iov[0].iov_base=buf;
  iov[0].iov_len =2048;
  tovalid = 0;

  bzero(&mh, sizeof(struct msghdr));
  bzero(buf, 2048);

  mh.msg_name    = &from;
  mh.msg_namelen = sizeof(struct sockaddr_in);
  mh.msg_iov     = iov;
  mh.msg_iovlen  = 1;
  mh.msg_control = address;
  mh.msg_controllen=sizeof(address);
  mh.msg_flags   = 0;

  if((msglen = recvmsg(ti->udpfd, &mh, 0)) == -1) {
    perror("recvmsg udp");
  }

  if(mh.msg_controllen > 0) {
    for(cmsg=CMSG_FIRSTHDR(&mh); cmsg; cmsg=CMSG_NXTHDR(&mh, cmsg)) {
      void *cdata = CMSG_DATA(cmsg);

      if(cmsg->cmsg_type == IP_RECVDSTADDR) {
	struct in_addr *i = cdata;

	to.sin_addr = *i;
	tovalid = 1;
      }
    }
  }

#if 0
  if(mh.msg_namelen == sizeof(struct sockaddr_in)*2) {
    /* must be hacked kernel, so to address is there */
    tovalid = 1;
  }
#endif

  if(ti->debug) {
    printf("Got an datagram (len=%d) from %s:%d ",
	   msglen,
	   inet_ntop(AF_INET, &from.sin_addr, addrfrom, sizeof(addrfrom)),
	   ntohs(from.sin_port));
    if(tovalid) {
      printf("to %s:%d ",
	     inet_ntop(AF_INET, &to.sin_addr, addrto, sizeof(addrto)),
	     ntohs(to.sin_port));
    } else {
      printf("to not valid    ");
    }
  }

  if(msglen >=1 && buf[0]==0x11) {
    /* this is a request to please echo things */
    /* format up the reply */

    if(ti->debug) {
      printf("echo request\n");
    }

    buf[0]=0x12;
    buf[1]=0;
    bcopy((caddr_t)&from.sin_port, &buf[2], 2);
    bcopy((caddr_t)&from.sin_addr, &buf[4], 4);
    
    if(sendto(ti->udpfd, buf, 8, 0,
	      (struct sockaddr *)&from, from.sin_len) == -1) {
      perror("reply to echo");
    }
    return;
  }

  if(msglen >=1 && buf[0]==0x12) {
    /* this must be a reply */

    if(ti->debug) {
      printf("echo reply\n");
    }
    
    bcopy(&buf[2], &translatedport, 2);
    translatedport = ntohs(translatedport);
    
    if(translatedport != UDP6TO4PORT || ti->meserver) {
      /* AhA! There is a NAT around, or we are a server */

      bcopy(&buf[4], &translatedaddr, 4);
      
      printf("Externally, I think that my port is %s:%u\n",
	     inet_ntop(AF_INET, &translatedaddr, addrto, sizeof(addrto)),
	     translatedport);
      
      ipdst = (unsigned char *)&ti->echoserver.s_addr;
      
      snprintf(sysbuf, 256,
	       "ifconfig tun%d inet6 2003:%02x%02x:%02x%02x:%02x%02x::%04x prefixlen 16 ",
	       ti->tunnumber,
	       buf[4], buf[5], buf[6], buf[7],
	       buf[2], buf[3],
	       ti->tunnumber+1,
	       ipdst[0],ipdst[1],ipdst[2],ipdst[3]);
      
      printf("Running %s\n", sysbuf);
      system(sysbuf);
      
      snprintf(sysbuf, 256,
	       "route add -net -inet6 :: 2003:%02x%02x:%02x%02x:2003::1",
	       ipdst[0],ipdst[1],ipdst[2],ipdst[3]);
      printf("Running %s\n", sysbuf);
      system(sysbuf);
    }

    return;
  }


  if(msglen >= 1 && (buf[0]&0xf0) == 0x60) {
    /* found an IPv6 packet. send to tunnel */

    if(write(ti->tunfd, buf, msglen) == -1) {
      perror("tunnel write");
    }
    return;
  }

  printf("Weird UDP packet of length %u with buf[0]=%02x\n",
	 msglen, buf[0]);
  return;
}

mainloop(struct tuninfo *ti)
{
  int done, pollresult;
  struct pollfd events[2];

  events[0].fd = ti->tunfd;
  events[0].events = POLLIN;
  events[0].revents = 0;

  events[1].fd = ti->udpfd;
  events[1].events = POLLIN;
  events[1].revents = 0;
  
  done = 0;

  if(debug) {
    printf("Entering mainloop\n");
  }

  while(!done) {
    while((pollresult = poll(events, 2, 500000)) > 0) {

      if(debug) {
	printf("poll result %d\n", pollresult);
      }

      if(events[0].revents) {
	tunprocess(ti);
      }
      
      if(events[1].revents) {
	udpprocess(ti);
      }
      
      events[0].revents = events[1].revents = 0;
    }
    if(pollresult == 0) {
      /* there was a timeout... do something */
    }
  }
}

usage()
{
  fprintf(stderr, "6to4tun [-e echoserver] [-t tun0device]\n");
}

main(int argc, char **argv)
{
  int opt;
  char *foo;
  struct tuninfo ti;
  struct hostent *he;
  int gotecho;
  char hostbuf[256];

  bzero(&ti, sizeof(struct tuninfo));
  gotecho = 0;

  while((opt=getopt(argc, argv, "de:m:st:")) != EOF) {
    switch(opt) {
    case 'd':
      ti.debug++;
      break;

    case 't':
      ti.tunnumber = strtoul(optarg, &foo, 0);
      if(optarg == foo) {
	fprintf(stderr, "Argument to -t should be a tunnel nunber not %u\n",
		optarg);
	exit(1);
      }
      break;

    case 's':
      ti.meserver++;
      break;

    case 'e':
      if(inet_pton(AF_INET, optarg, &ti.echoserver) == 1) {
	/* okay done. */
	gotecho++;
	break;
      }

      he = gethostbyname2(optarg, AF_INET);
      if(he == NULL) {
	fprintf(stderr, "Can not decode %s as either IPv4 address or hostname\n",
		optarg);
	herror(optarg);
	exit(1);
      }
      if(he->h_length > sizeof(struct in_addr)) {
	printf("Multiple addresses returned, using %s\n",
	       inet_ntop(AF_INET, he->h_addr, hostbuf, sizeof(hostbuf)));
      }

      bcopy(he->h_addr, &ti.echoserver, sizeof(struct in_addr));
      gotecho++;
      break;

    case 'm':
      if(inet_pton(AF_INET, optarg, &ti.myaddr) == 1) {
	/* okay done. */
	ti.myaddr_set++;
	break;
      }

      fprintf(stderr, "Can not decode %s as IPv4 address\n",
	      optarg);
      exit(1);
      break;

    case '?':
      usage();
      exit(1);
    }
  }

  if(!gotecho && !ti.meserver) {
    usage();
    exit(1);
  }

  if(ti.debug) {
    printf("Debug level set to %u\n", ti.debug);
  }

  tuninit(&ti);

  udpinit(&ti);

  send_echorequest(&ti);

  mainloop(&ti);
  
}
  
